#include <stdio.h>
#include <assert.h>
#include <goatvr.h>
#include "app.h"
#include "opengl.h"
#include "sdr.h"
#include "texture.h"
#include "mesh.h"
#include "meshgen.h"
#include "scene.h"
#include "metascene.h"
#include "datamap.h"
#include "ui.h"
#include "opt.h"
#include "post.h"
#include "renderer.h"
#include "avatar.h"
#include "vrinput.h"
#include "exman.h"
#include "blob_exhibit.h"
#include "dbg_gui.h"

#define NEAR_CLIP	5.0
#define FAR_CLIP	10000.0

static void draw_scene();
static void toggle_flight();
static void calc_framerate();
static Ray calc_pick_ray(int x, int y, const Mat4 &view_mat, const Mat4 &proj_mat);

long time_msec;
int win_width, win_height;
float win_aspect;
bool fb_srgb;
bool opt_gear_wireframe;

TextureSet texman;
SceneSet sceneman;

unsigned int sdr_ltmap, sdr_ltmap_notex;

static Avatar avatar;

static float cam_dist = 0.0;
static float floor_y;	// last floor height
static float user_eye_height = 165;

static float walk_speed = 300.0f;
static float mouse_speed = 0.5f;
static bool show_walk_mesh, noclip = false;

static bool have_headtracking, have_handtracking, should_swap;

static int prev_mx, prev_my;
static bool bnstate[8];
static bool keystate[256];
static bool gpad_bnstate[64];
static Vec2 joy_move, joy_look;
static float joy_deadzone = 0.01;

static float framerate;

static Mat4 view_matrix, mouse_view_matrix, proj_matrix;
static MetaScene *mscn;
static unsigned int sdr_post_gamma;

static long prev_msec;

static ExhibitManager *exman;
static BlobExhibit *blobs;
static bool show_blobs;

static ExSelection ex_sel;

static Renderer *rend;


bool app_init(int argc, char **argv)
{
	set_log_file("demo.log");

	if(init_opengl() == -1) {
		return false;
	}

	if(!init_options(argc, argv, "demo.conf")) {
		return false;
	}
	app_resize(opt.width, opt.height);
	app_fullscreen(opt.fullscreen);

	if(opt.vr) {
		if(goatvr_init() == -1) {
			return false;
		}
		goatvr_set_origin_mode(GOATVR_HEAD);
		goatvr_set_units_scale(100.0f);

		goatvr_startvr();
		should_swap = goatvr_should_swap() != 0;
		user_eye_height = goatvr_get_eye_height();
		have_headtracking = goatvr_have_headtracking();
		have_handtracking = goatvr_have_handtracking();

		goatvr_recenter();
	}

	if(fb_srgb) {
		int srgb_capable;
		glGetIntegerv(GL_FRAMEBUFFER_SRGB_CAPABLE_EXT, &srgb_capable);
		printf("Framebuffer %s sRGB-capable\n", srgb_capable ? "is" : "is not");
		if(srgb_capable) {
			glEnable(GL_FRAMEBUFFER_SRGB);
		} else {
			fb_srgb = 0;
		}
	}

	glEnable(GL_MULTISAMPLE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);

	if(!init_debug_gui()) {
		return false;
	}

	Mesh::use_custom_sdr_attr = false;

	float ambient[] = {0.0, 0.0, 0.0, 0.0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	glClearColor(1, 1, 1, 1);

	init_audio();

	if(!init_vrhands()) {
		return false;
	}

	mscn = new MetaScene;
	if(!mscn->load(opt.scenefile ? opt.scenefile : "data/museum.scene")) {
		return false;
	}

	avatar.pos = mscn->start_pos;
	Vec3 dir = rotate(Vec3(0, 0, 1), mscn->start_rot);
	dir.y = 0;
	avatar.body_rot = rad_to_deg(acos(dot(dir, Vec3(0, 0, 1))));

	exman = new ExhibitManager;
	/*
	if(!exman->load(mscn, "data/exhibits")) {
		//return false;
	}
	*/

	blobs = new BlobExhibit;
	blobs->node = new SceneNode;
	blobs->init();
	blobs->node->set_position(Vec3(-680, 160, -100));
	blobs->node->set_scaling(Vec3(28, 28, 28));
	blobs->node->update(0);

	exman->add(blobs);

	if(!(sdr_ltmap_notex = create_program_load("sdr/lightmap.v.glsl", "sdr/lightmap-notex.p.glsl"))) {
		return false;
	}
	set_uniform_int(sdr_ltmap_notex, "texmap", MTL_TEX_DIFFUSE);
	set_uniform_int(sdr_ltmap_notex, "lightmap", MTL_TEX_LIGHTMAP);

	if(!(sdr_ltmap = create_program_load("sdr/lightmap.v.glsl", "sdr/lightmap-tex.p.glsl"))) {
		return false;
	}
	set_uniform_int(sdr_ltmap, "texmap", MTL_TEX_DIFFUSE);
	set_uniform_int(sdr_ltmap, "lightmap", MTL_TEX_LIGHTMAP);

	if(!fb_srgb) {
		sdr_post_gamma = create_program_load("sdr/post_gamma.v.glsl", "sdr/post_gamma.p.glsl");
	}

	rend = new Renderer;
	rend->set_scene(mscn);

	glUseProgram(0);

	if(opt.vr || opt.fullscreen) {
		app_grab_mouse(true);
	}

	if(mscn->music && opt.music) {
		mscn->music->play(AUDIO_PLAYMODE_LOOP);
	}
	return true;
}

void app_cleanup()
{
	if(mscn->music) {
		mscn->music->stop();
	}
	destroy_audio();

	app_grab_mouse(false);
	if(opt.vr) {
		goatvr_shutdown();
	}
	destroy_vrhands();

	delete rend;

	delete exman;

	texman.clear();
	sceneman.clear();

	cleanup_debug_gui();
}

static bool constrain_walk_mesh(const Vec3 &v, Vec3 *newv)
{
	Mesh *wm = mscn->walk_mesh;
	if(!wm) {
		*newv = v;
		return true;
	}

	Ray downray = Ray(v, Vec3(0, -1, 0));
	HitPoint hit;
	if(mscn->walk_mesh->intersect(downray, &hit)) {
		*newv = hit.pos;
		newv->y += user_eye_height;
		return true;
	}
	return false;
}

static void update(float dt)
{
	texman.update();
	sceneman.update();

	mscn->update(dt);
	exman->update(dt);

	float speed = walk_speed * dt;
	Vec3 dir;

	// joystick
	float jdeadsq = joy_deadzone * joy_deadzone;
	float jmove_lensq = length_sq(joy_move);
	float jlook_lensq = length_sq(joy_look);

	if(jmove_lensq > jdeadsq) {
		float len = sqrt(jmove_lensq);
		jmove_lensq -= jdeadsq;

		float mag = len * len;
		dir.x += mag * joy_move.x / len * 2.0 * speed;
		dir.z += mag * joy_move.y / len * 2.0 * speed;
	}
	if(jlook_lensq > jdeadsq) {
		float len = sqrt(jlook_lensq);
		jlook_lensq -= jdeadsq;

		float mag = len * len;
		avatar.body_rot += mag * joy_look.x / len * 200.0 * dt;
		avatar.head_alt += mag * joy_look.y / len * 100.0 * dt;
		if(avatar.head_alt < -90.0f) avatar.head_alt = -90.0f;
		if(avatar.head_alt > 90.0f) avatar.head_alt = 90.0f;
	}

	// keyboard move
	if(keystate[(int)'w']) {
		dir.z -= speed;
	}
	if(keystate[(int)'s']) {
		dir.z += speed;
	}
	if(keystate[(int)'d']) {
		dir.x += speed;
	}
	if(keystate[(int)'a']) {
		dir.x -= speed;
	}
	if(keystate[(int)'q'] || gpad_bnstate[GPAD_UP]) {
		avatar.pos.y += speed;
	}
	if(keystate[(int)'z'] || gpad_bnstate[GPAD_DOWN]) {
		avatar.pos.y -= speed;
	}

	float theta = M_PI * avatar.body_rot / 180.0f;
	Vec3 newpos;
	newpos.x = avatar.pos.x + cos(theta) * dir.x - sin(theta) * dir.z;
	newpos.y = avatar.pos.y;
	newpos.z = avatar.pos.z + sin(theta) * dir.x + cos(theta) * dir.z;

	if(noclip) {
		avatar.pos = newpos;
	} else {
		if(!constrain_walk_mesh(newpos, &avatar.pos)) {
			float dtheta = M_PI / 32.0;
			float theta = dtheta;
			Vec2 dir2d = newpos.xz() - avatar.pos.xz();

			for(int i=0; i<16; i++) {
				Vec2 dvec = rotate(dir2d, theta);
				Vec3 pos = avatar.pos + Vec3(dvec.x, 0, dvec.y);
				if(constrain_walk_mesh(pos, &avatar.pos)) {
					break;
				}
				dvec = rotate(dir2d, -theta);
				pos = avatar.pos + Vec3(dvec.x, 0, dvec.y);
				if(constrain_walk_mesh(pos, &avatar.pos)) {
					break;
				}
				theta += dtheta;
			}
		}
		floor_y = avatar.pos.y - user_eye_height;
	}

	// TODO move to avatar
	// calculate mouselook view matrix
	mouse_view_matrix = Mat4::identity;
	mouse_view_matrix.pre_translate(0, 0, -cam_dist);
	if(!have_headtracking) {
		mouse_view_matrix.pre_rotate_x(deg_to_rad(avatar.head_alt));
	}
	mouse_view_matrix.pre_rotate_y(deg_to_rad(avatar.body_rot));
	mouse_view_matrix.pre_translate(-avatar.pos.x, -avatar.pos.y, -avatar.pos.z);

	// update hand-tracking
	if(have_handtracking) {
		update_vrhands(&avatar);
	}
}

static void set_light(int idx, const Vec3 &pos, const Vec3 &color)
{
	unsigned int lt = GL_LIGHT0 + idx;
	float posv[] = { pos.x, pos.y, pos.z, 1 };
	float colv[] = { color.x, color.y, color.z, 1 };

	glEnable(lt);
	glLightfv(lt, GL_POSITION, posv);
	glLightfv(lt, GL_DIFFUSE, colv);
	glLightfv(lt, GL_SPECULAR, colv);
}

void app_display()
{
	float dt = (float)(time_msec - prev_msec) / 1000.0f;
	prev_msec = time_msec;

	if(debug_gui) {
		ImGui::GetIOPtr()->DeltaTime = dt;
		ImGui::NewFrame();

		ImGui::ShowTestWindow();
	}

	if(opt.vr) {
		// VR mode
		goatvr_draw_start();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		update(dt);

		for(int i=0; i<2; i++) {
			// for each eye
			goatvr_draw_eye(i);

			proj_matrix = goatvr_projection_matrix(i, NEAR_CLIP, FAR_CLIP);
			glMatrixMode(GL_PROJECTION);
			glLoadMatrixf(proj_matrix[0]);

			view_matrix = mouse_view_matrix * Mat4(goatvr_view_matrix(i));
			glMatrixMode(GL_MODELVIEW);
			glLoadMatrixf(view_matrix[0]);

			draw_scene();
			if(have_handtracking) {
				draw_vrhands();
			}

			if(debug_gui) {
				ImGui::Render();
			}
		}
		goatvr_draw_done();

		if(should_swap) {
			app_swap_buffers();
		}

	} else {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		update(dt);

		proj_matrix.perspective(deg_to_rad(50.0), win_aspect, NEAR_CLIP, FAR_CLIP);
		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(proj_matrix[0]);

		view_matrix = mouse_view_matrix;
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(view_matrix[0]);

		draw_scene();

		if(!fb_srgb && sdr_post_gamma) {
			slow_post(sdr_post_gamma);
			glUseProgram(0);
		}

		if(debug_gui) {
			ImGui::Render();
		}
		app_swap_buffers();
	}
	assert(glGetError() == GL_NO_ERROR);

	calc_framerate();
}


static void draw_scene()
{
	static const Vec3 lpos[] = { Vec3(-50, 75, 100), Vec3(100, 0, 30), Vec3(-10, -10, 60) };
	set_light(0, lpos[0], Vec3(1.0, 0.8, 0.7) * 0.8);
	set_light(1, lpos[1], Vec3(0.6, 0.7, 1.0) * 0.6);
	set_light(2, lpos[2], Vec3(0.8, 1.0, 0.8) * 0.3);

	rend->draw();
	exman->draw();

	/*
	if(have_handtracking) {
		Mat4 head_xform = inverse(mouse_view_matrix);//goatvr_head_matrix();
		Mat4 head_dir_xform = head_xform.upper3x3();

		glUseProgram(0);
		glPushAttrib(GL_ENABLE_BIT);
		glDisable(GL_LIGHTING);
		glBegin(GL_LINES);
		for(int i=0; i<2; i++) {
			if(hand[i].valid) {
				glColor3f(i, 1 - i, i);
			} else {
				glColor3f(0.5, 0.5, 0.5);
			}
			Vec3 v = head_xform * hand[i].pos;
			Vec3 dir = head_dir_xform * rotate(Vec3(0, 0, -1), hand[i].rot) * 20.0f;
			Vec3 up = head_dir_xform * rotate(Vec3(0, 1, 0), hand[i].rot) * 10.0f;
			Vec3 right = head_dir_xform * rotate(Vec3(1, 0, 0), hand[i].rot) * 10.0f;

			glVertex3f(v.x, v.y, v.z);
			glVertex3f(v.x + dir.x, v.y + dir.y, v.z + dir.z);
			glVertex3f(v.x - right.x, v.y - right.y, v.z - right.z);
			glVertex3f(v.x + right.x, v.y + right.y, v.z + right.z);
			glVertex3f(v.x - up.x, v.y - up.y, v.z - up.z);
			glVertex3f(v.x + up.x, v.y + up.y, v.z + up.z);
		}
		glEnd();
		glPopAttrib();
	}
	*/

	if(show_walk_mesh && mscn->walk_mesh) {
		glPushAttrib(GL_ENABLE_BIT);
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);
		glDisable(GL_LIGHTING);
		glEnable(GL_POLYGON_OFFSET_FILL);

		glUseProgram(0);

		glPolygonOffset(-1, 1);
		glDepthMask(0);

		glColor3f(0.3, 0.08, 0.01);
		mscn->walk_mesh->draw();

		glDepthMask(1);

		glPopAttrib();
	}

	print_text(Vec2(9 * win_width / 10, 20), Vec3(1, 1, 0), "fps: %.1f", framerate);
	draw_ui();
}


void app_reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	goatvr_set_fb_size(x, y, 1.0f);
	debug_gui_reshape(x, y);
}

void app_keyboard(int key, bool pressed)
{
	unsigned int mod = app_get_modifiers();

	if(debug_gui && !(pressed && (key == '`' || key == 27))) {
		debug_gui_key(key, pressed, mod);
		return;	// ignore all keystrokes when GUI is visible
	}

	if(pressed) {
		switch(key) {
		case 27:
			app_quit();
			break;

		case '\n':
		case '\r':
			if(mod & MOD_ALT) {
				app_toggle_fullscreen();
			}
			break;

		case '`':
			debug_gui = !debug_gui;
			show_message("debug gui %s", debug_gui ? "enabled" : "disabled");
			break;

		case 'm':
			app_toggle_grab_mouse();
			show_message("mouse %s", app_is_mouse_grabbed() ? "grabbed" : "released");
			break;

		case 'w':
			if(mod & MOD_CTRL) {
				show_walk_mesh = !show_walk_mesh;
				show_message("walk mesh: %s", show_walk_mesh ? "on" : "off");
			}
			break;

		case 'c':
			if(mod & MOD_CTRL) {
				noclip = !noclip;
				show_message(noclip ? "no clip" : "clip");
			}
			break;

		case 'f':
			toggle_flight();
			break;

		case 'p':
			if(mod & MOD_CTRL) {
				fb_srgb = !fb_srgb;
				show_message("gamma correction for non-sRGB framebuffers: %s\n", fb_srgb ? "off" : "on");
			}
			break;

		case '=':
			walk_speed *= 1.25;
			show_message("walk speed: %g", walk_speed);
			break;

		case '-':
			walk_speed *= 0.75;
			show_message("walk speed: %g", walk_speed);
			break;

		case ']':
			mouse_speed *= 1.2;
			show_message("mouse speed: %g", mouse_speed);
			break;

		case '[':
			mouse_speed *= 0.8;
			show_message("mouse speed: %g", mouse_speed);
			break;

		case 'b':
			show_blobs = !show_blobs;
			show_message("blobs: %s\n", show_blobs ? "on" : "off");
			break;

		case ' ':
			goatvr_recenter();
			show_message("VR recenter\n");
			break;

		case 'x':
			exman->load(mscn, "data/exhibits");
			break;
		}
	}

	if(key < 256 && !(mod & (MOD_CTRL | MOD_ALT))) {
		keystate[key] = pressed;
	}
}

void app_mouse_button(int bn, bool pressed, int x, int y)
{
	if(debug_gui) {
		debug_gui_mbutton(bn, pressed, x, y);
		return;	// ignore mouse events while GUI is visible
	}

	prev_mx = x;
	prev_my = y;
	bnstate[bn] = pressed;

	if(bn == 0) {
		if(!pressed) {
			if(ex_sel.ex) {
				ex_sel.ex = 0;
			} else {
				Ray ray = calc_pick_ray(x, y, view_matrix, proj_matrix);
				ex_sel = exman->select(ray);
			}
		}
	}
}

static inline void mouse_look(float dx, float dy)
{
	float scrsz = (float)win_height;
	avatar.body_rot += dx * 512.0 / scrsz;
	avatar.head_alt += dy * 512.0 / scrsz;

	if(avatar.head_alt < -90) avatar.head_alt = -90;
	if(avatar.head_alt > 90) avatar.head_alt = 90;
}

static void mouse_zoom(float dx, float dy)
{
	cam_dist += dy * 0.1;
	if(cam_dist < 0.0) cam_dist = 0.0;
}

void app_mouse_motion(int x, int y)
{
	if(debug_gui) {
		debug_gui_mmotion(x, y);
		return;	// ignore mouse events while GUI is visible
	}

	int dx = x - prev_mx;
	int dy = y - prev_my;
	prev_mx = x;
	prev_my = y;

	if(!dx && !dy) return;

	if(ex_sel.ex) {
		Vec3 pos = ex_sel.ex->node->get_node_position();
		Vec3 dir = transpose(view_matrix.upper3x3()) * Vec3(dx * 1.0, dy * -1.0, 0);

		ex_sel.ex->node->set_position(pos + dir);
	}

	if(bnstate[2]) {
		mouse_look(dx, dy);
	}
}

void app_mouse_delta(int dx, int dy)
{
	if(bnstate[2]) {
		mouse_zoom(dx * mouse_speed, dy * mouse_speed);
	} else {
		mouse_look(dx * mouse_speed, dy * mouse_speed);
	}
}

void app_mouse_wheel(int dir)
{
	if(debug_gui) {
		debug_gui_wheel(dir);
	}
}

void app_gamepad_axis(int axis, float val)
{
	switch(axis) {
	case 0:
		joy_move.x = val;
		break;
	case 1:
		joy_move.y = val;
		break;

	case 2:
		joy_look.x = val;
		break;
	case 3:
		joy_look.y = val;
		break;
	}
}

void app_gamepad_button(int bn, bool pressed)
{
	gpad_bnstate[bn] = pressed;

	if(pressed) {
		switch(bn) {
		case GPAD_LSTICK:
			toggle_flight();
			break;

		case GPAD_X:
			show_blobs = !show_blobs;
			show_message("blobs: %s\n", show_blobs ? "on" : "off");
			break;

		case GPAD_START:
			goatvr_recenter();
			show_message("VR recenter\n");
			break;

		default:
			break;
		}
	}
}

static void toggle_flight()
{
	static float prev_walk_speed = -1.0;
	if(prev_walk_speed < 0.0) {
		noclip = true;
		prev_walk_speed = walk_speed;
		walk_speed = 1000.0;
		show_message("fly mode\n");
	} else {
		noclip = false;
		walk_speed = prev_walk_speed;
		prev_walk_speed = -1.0;
		show_message("walk mode\n");
	}
}

static void calc_framerate()
{
	static int ncalc;
	static int nframes;
	static long prev_upd;

	long elapsed = time_msec - prev_upd;
	if(elapsed >= 1000) {
		framerate = (float)nframes / (float)(elapsed * 0.001);
		nframes = 1;
		prev_upd = time_msec;

		/*if(++ncalc >= 5) {
			printf("fps: %f\n", framerate);
			ncalc = 0;
		}*/
	} else {
		++nframes;
	}
}

static Ray calc_pick_ray(int x, int y, const Mat4 &view_mat, const Mat4 &proj_mat)
{
	return Ray();	// TODO
}
