/*! \file
 * \brief Metascene implementation
 *
 * Loading starts at `MetaScene::load`, which calls `ts_load` (libtreestore)
 * to load the metascene description tree into memory. Then `proc_node` is
 * called at the root to recursively process the tree. `scenefile` nodes are
 * handed over to `proc_scenefile`, which will trigger scene loading for any
 * nodes with a `file` attribute. Scene loading is handled by requesting the
 * filename from the scene resource manager, which is of type [SceneSet](\ref SceneSet).
 */
#include <assert.h>
#include <string>
#include <regex>
#include "metascene.h"
#include "scene.h"
#include "treestore.h"
#include "logger.h"
#include "app.h"
#include "dbg_gui.h"

#if defined(WIN32) || defined(__WIN32__)
#include <malloc.h>
#else
#include <alloca.h>
#endif

struct MaterialEdit {
	std::regex name_re;
	int attr;
	Texture *tex;
};

static bool proc_node(MetaScene *mscn, struct ts_node *node);
static bool proc_scenefile(MetaScene *mscn, struct ts_node *node);
static bool proc_mtledit(MetaScene *mscn, MaterialEdit *med, struct ts_node *node);
static bool proc_music(MetaScene *mscn, struct ts_node *node);
static void apply_mtledit(Scene *scn, const MaterialEdit &med);
static void apply_mtledit(Material *mtl, const MaterialEdit &med);
static struct ts_attr *attr_inscope(struct ts_node *node, const char *name);

static void print_scene_graph(SceneNode *n, int level);


MetaScene::MetaScene()
{
	walk_mesh = 0;
	music = 0;
}

MetaScene::~MetaScene()
{
	delete walk_mesh;
	delete music;
}

bool MetaScene::load(const char *fname)
{
	struct ts_node *root = ts_load(fname);
	if(!root || strcmp(root->name, "scene") != 0) {
		ts_free_tree(root);
		error_log("failed to load scene metadata: %s\n", fname);
		return false;
	}

	bool res = proc_node(this, root);
	ts_free_tree(root);

	/*info_log("loaded scene: %s\n", fname);
	info_log("scene graph:\n");
	print_scene_graph(scn->nodes, 0);
	*/
	return res;
}

void MetaScene::update(float dt)
{
	bool expanded;
	static char text[256];
	if(debug_gui) {
		ImGui::Begin("MetaScene nodes", 0, 0);
	}

	int nscn = scenes.size();
	for(int i=0; i<nscn; i++) {

		if(debug_gui) {
			if(scenes[i]->name.empty()) {
				sprintf(text, "scene %3d", i);
			} else {
				sprintf(text, "scene %3d: %s", i, scenes[i]->name.c_str());
			}
			expanded = parent_expanded = ImGui::TreeNode(text);
		}

		scenes[i]->update(dt);

		if(debug_gui && expanded) {
			ImGui::TreePop();
		}
	}

	if(debug_gui) {
		ImGui::End();
	}
}

void MetaScene::draw() const
{
	int nscn = scenes.size();
	for(int i=0; i<nscn; i++) {
		scenes[i]->draw();
	}
}

SceneNode *MetaScene::find_node(const char *name) const
{
	int num = scenes.size();
	for(int i=0; i<num; i++) {
		SceneNode *n = scenes[i]->find_node(name);
		if(n) return n;
	}
	return 0;
}

SceneNode *MetaScene::match_node(const char *qstr) const
{
	int num = scenes.size();
	for(int i=0; i<num; i++) {
		SceneNode *n = scenes[i]->match_node(qstr);
		if(n) return n;
	}
	return 0;
}

std::list<SceneNode*> MetaScene::match_nodes(const char *qstr) const
{
	std::list<SceneNode*> res;
	int num = scenes.size();
	for(int i=0; i<num; i++) {
		std::list<SceneNode*> tmp = scenes[i]->match_nodes(qstr);
		if(!tmp.empty()) {
			res.splice(res.end(), tmp);
		}
	}
	return std::move(res);
}

Scene *MetaScene::extract_nodes(const char *qstr)
{
	Scene *scn = 0;
	int nscn = scenes.size();
	for(int i=0; i<nscn; i++) {
		Scene *tmp = scenes[i]->extract_nodes(qstr);
		if(tmp) {
			if(!scn) {
				scn = tmp;
			} else {
				scn->merge(tmp);
				delete tmp;
			}
		}
	}
	return scn;
}

static bool proc_node(MetaScene *mscn, struct ts_node *node)
{
	struct ts_node *c = node->child_list;
	while(c) {
		if(!proc_node(mscn, c)) {
			return false;
		}
		c = c->next;
	}

	// do this last to allow other contents of the node to do their thing
	if(strcmp(node->name, "scenefile") == 0) {
		return proc_scenefile(mscn, node);

	} else if(strcmp(node->name, "remap") == 0) {
		const char *match = ts_get_attr_str(node, "match");
		const char *replace = ts_get_attr_str(node, "replace");
		if(match && replace) {
			mscn->datamap.map(match, replace);
		}

	} else if(strcmp(node->name, "music") == 0) {
		return proc_music(mscn, node);
	}

	return true;
}



struct SceneData {
	MetaScene *meta;
	std::string walkmesh_regexp, spawn_regexp;
	std::vector<MaterialEdit> mtledit;
};

/*! Processes a `scenefile` node. And kicks off scene loading (if necessary) by
 * calling `SceneSet::get`.
 */
static bool proc_scenefile(MetaScene *mscn, struct ts_node *node)
{
	const char *fname = ts_get_attr_str(node, "file");
	if(fname) {
		SceneData *sdat = new SceneData;
		sdat->meta = mscn;

		// datapath
		struct ts_attr *adpath = attr_inscope(node, "datapath");
		if(adpath && adpath->val.type == TS_STRING) {
			mscn->datamap.set_path(adpath->val.str);
		}

		// strip path
		struct ts_attr *aspath = attr_inscope(node, "strip_path");
		if(aspath && aspath->val.type == TS_NUMBER) {
			mscn->datamap.set_strip(aspath->val.inum);
		}

		// walkmesh
		struct ts_attr *awmesh = attr_inscope(node, "walkmesh");
		if(awmesh && awmesh->val.type == TS_STRING) {
			sdat->walkmesh_regexp = std::string(awmesh->val.str);
		}

		// spawn node
		struct ts_attr *awspawn = attr_inscope(node, "spawn");
		if(awspawn) {
			switch(awspawn->val.type) {
			case TS_VECTOR:
				mscn->start_pos = Vec3(awspawn->val.vec[0], awspawn->val.vec[1],
						awspawn->val.vec[2]);
				break;

			case TS_STRING:
			default:
				sdat->spawn_regexp = std::string(awspawn->val.str);
			}
		}
		if((awspawn = attr_inscope(node, "spawn_rot")) && awspawn->val.type == TS_VECTOR) {
			Quat rot;
			rot.rotate(Vec3(1, 0, 0), deg_to_rad(awspawn->val.vec[0]));
			rot.rotate(Vec3(0, 1, 0), deg_to_rad(awspawn->val.vec[1]));
			rot.rotate(Vec3(0, 0, 1), deg_to_rad(awspawn->val.vec[2]));
			mscn->start_rot = rot;
		}

		int namesz = mscn->datamap.lookup(fname, 0, 0);
		char *namebuf = (char*)alloca(namesz + 1);
		if(mscn->datamap.lookup(fname, namebuf, namesz + 1)) {
			fname = namebuf;
		}

		// material edits are kept in a list to be applied when the scene has been loaded
		struct ts_node *child = node->child_list;
		while(child) {
			MaterialEdit medit;
			if(proc_mtledit(mscn, &medit, child)) {
				sdat->mtledit.push_back(medit);
			}
			child = child->next;
		}

		Scene *newscn = sceneman.get(fname);
		/* NOTE: setting all these after get() is not a race condition, because
		 * scene_loaded() which uses this, will only run in our main loop during
		 * SceneSet::update() on the main thread.
		 */
		newscn->datamap = mscn->datamap;
		mscn->datamap.clear();

		newscn->metascn = mscn;
		mscn->scndata[newscn] = sdat;
	}
	return true;
}

bool MetaScene::scene_loaded(Scene *newscn)
{
	SceneData *sdat = (SceneData*)scndata[newscn];
	if(!sdat) {
		error_log("MetaScene::scene_loaded called, but corresponding SceneData not found\n");
		return false;
	}

	// extract the walk mesh if necessary
	Scene *wscn;
	if(!sdat->walkmesh_regexp.empty() && (wscn = newscn->extract_nodes(sdat->walkmesh_regexp.c_str()))) {
		// apply all transformations to the meshes
		wscn->apply_xform();

		int nmeshes = wscn->meshes.size();
		for(int i=0; i<nmeshes; i++) {
			Mesh *m = wscn->meshes[i];

			if(walk_mesh) {
				walk_mesh->append(*m);
			} else {
				walk_mesh = m;
				wscn->remove_mesh(m);	// to save it from destruction
			}
		}

		delete wscn;
	}

	// extract the spawn node
	if(!sdat->spawn_regexp.empty() && (wscn = newscn->extract_nodes(sdat->spawn_regexp.c_str()))) {

		int nmeshes = wscn->meshes.size();
		int nnodes = wscn->nodes ? wscn->nodes->get_num_children() : 0;

		if(nmeshes) {
			Vec3 pos;
			for(int i=0; i<nmeshes; i++) {
				const Sphere &bsph = wscn->meshes[i]->get_bsphere();
				pos += bsph.center;
			}
			pos /= (float)nmeshes;
			sdat->meta->start_pos = pos;

		} else if(nnodes) {
			// just use the first one
			SceneNode *first = wscn->nodes->get_child(0);
			sdat->meta->start_pos = first->get_position();
			sdat->meta->start_rot = first->get_rotation();
		}
		delete wscn;
	}

	int num_medits = sdat->mtledit.size();
	for(int i=0; i<num_medits; i++) {
		// perform material edits
		apply_mtledit(newscn, sdat->mtledit[i]);
	}

	scenes.push_back(newscn);
	return true;
}

static bool proc_mtledit(MetaScene *mscn, MaterialEdit *med, struct ts_node *node)
{
	if(strcmp(node->name, "mtledit") != 0) {
		return false;
	}

	const char *restr = ".*";
	struct ts_attr *amtl = ts_get_attr(node, "material");
	if(amtl && amtl->val.type == TS_STRING) {
		restr = amtl->val.str;
	}

	med->name_re = std::regex(restr);

	node = node->child_list;
	while(node) {
		struct ts_node *cn = node;
		node = node->next;

		if(strcmp(cn->name, "texture") == 0) {
			// add/change/remove a texture
			struct ts_attr *atype = ts_get_attr(cn, "type");
			struct ts_attr *afile = ts_get_attr(cn, "file");

			int textype = MTL_TEX_DIFFUSE;
			if(atype) {
				if(atype->val.type == TS_STRING) {
					textype = mtl_parse_type(atype->val.str);
				} else if(atype->val.type == TS_NUMBER) {
					textype = atype->val.inum;
					if(textype < 0 || textype >= NUM_MTL_TEXTURES) {
						error_log("invalid texture in mtledit: %d\n", textype);
						continue;
					}
				} else {
					error_log("unexpected texture type in mtledit: %s\n", atype->val.str);
					continue;
				}
			}

			med->attr = textype;

			if(!afile || !afile->val.str || !*afile->val.str) {
				// remove
				med->tex = 0;
			} else {
				med->tex = texman.get_texture(afile->val.str, TEX_2D, &mscn->datamap);
			}
		}
		// TODO add more edit modes
	}

	return true;
}

static void apply_mtledit(Scene *scn, const MaterialEdit &med)
{
	// search all the objects to find matching material names
	int nobj = scn->objects.size();
	for(int i=0; i<nobj; i++) {
		Object *obj = scn->objects[i];
		if(std::regex_match(obj->mtl.name, med.name_re)) {
			apply_mtledit(&obj->mtl, med);
		}
	}
}

static bool proc_music(MetaScene *mscn, struct ts_node *node)
{
	const char *fname = ts_get_attr_str(node, "file");
	if(fname) {
		SceneData *sdat = new SceneData;
		sdat->meta = mscn;

		// datapath
		struct ts_attr *adpath = attr_inscope(node, "datapath");
		if(adpath && adpath->val.type == TS_STRING) {
			mscn->datamap.set_path(adpath->val.str);
		}

		int namesz = mscn->datamap.lookup(fname, 0, 0);
		char *namebuf = (char*)alloca(namesz + 1);
		if(mscn->datamap.lookup(fname, namebuf, namesz + 1)) {
			fname = namebuf;
		}

		OggVorbisStream *ovstream = new OggVorbisStream;
		if(!ovstream->open(fname)) {
			delete ovstream;
			return false;
		}

		delete mscn->music;
		mscn->music = ovstream;
	}
	return true;
}

static void apply_mtledit(Material *mtl, const MaterialEdit &med)
{
	// TODO more edit modes...
	if(med.tex) {
		mtl->add_texture(med.tex, med.attr);
	} else {
		Texture *tex = mtl->stdtex[med.attr];
		if(tex) {
			mtl->remove_texture(tex);
		}
	}
}

static struct ts_attr *attr_inscope(struct ts_node *node, const char *name)
{
	struct ts_attr *attr = 0;

	while(node && !(attr = ts_get_attr(node, name))) {
		node = node->parent;
	}
	return attr;
}

static void print_scene_graph(SceneNode *n, int level)
{
	if(!n) return;

	for(int i=0; i<level; i++) {
		info_log("  ");
	}

	int nobj = n->get_num_objects();
	if(nobj) {
		info_log("%s - %d obj\n", n->get_name(), n->get_num_objects());
	} else {
		info_log("%s\n", n->get_name());
	}

	for(int i=0; i<n->get_num_children(); i++) {
		print_scene_graph(n->get_child(i), level + 1);
	}
}
